<?php

  //--------------------------------------------------------------------------
  // PHPcoord
  // phpcoord.php
  //
  // (c) 2005 Jonathan Stott
  //
  // Created on 11-Aug-2005
  //
  // 0.1 - 11 Aug 2005
  //  - Initial version
  //--------------------------------------------------------------------------
  
  
  // ================================================================== LatLng
  
  class LatLng {
  
    var $lat;
    var $lng;
    
    function LatLng($lat, $lng) {
      $this->lat = $lat;
      $this->lng = $lng;
    }

    function toString() {
      return "(" . $this->lat . ", " . $this->lng . ")";
    }
    
  }
  
  
  // =================================================================== OSRef
  
  // References given with OSRef are accurate to 1m.
  class OSRef {
  
    var $easting;
    var $northing;
    
    function OSRef($easting, $northing) {
      $this->easting  = $easting;
      $this->northing = $northing;
    }
  
    function toString() {
      return "(" . $this->easting . ", " . $this->northing . ")";
    }
    
    function toSixFigureString() {
      $hundredkmE = floor($this->easting / 100000);
      $hundredkmN = floor($this->northing / 100000);
      $firstLetter = "";
      if ($hundredkmN < 5) {
        if ($hundredkmE < 5) {
          $firstLetter = "S";
        } else {
          $firstLetter = "T";
        }
      } else if ($hundredkmN < 10) {
        if ($hundredkmE < 5) {
          $firstLetter = "N";
        } else {
          $firstLetter = "O";
        }
      } else {
        $firstLetter = "H";
      }
      
      $secondLetter = "";
      $index = 65 + ((4 - ($hundredkmN % 5)) * 5) + ($hundredkmE % 5);
      $ti = $index;
      if ($index >= 73) $index++;
      $secondLetter = chr($index);
      
      $e = floor(($this->easting - (100000 * $hundredkmE)) / 100);
      $n = floor(($this->northing - (100000 * $hundredkmN)) / 100);
      $es = $e;
      if ($e < 100) $es = "0$es";
      if ($e < 10)  $es = "0$es";
      $ns = $n;
      if ($n < 100) $ns = "0$ns";
      if ($n < 10)  $ns = "0$ns";
      
      return $firstLetter . $secondLetter . $es . $ns;
    }
  
  }
  
  
  // ================================================================== UTMRef
  
  class UTMRef {
  
    var $easting;
    var $northing;
    var $latZone;
    var $lngZone;
    
    function UTMRef($easting, $northing, $latZone, $lngZone) {
      $this->easting  = $easting;
      $this->northing = $northing;
      $this->latZone  = $latZone;
      $this->lngZone  = $lngZone;
    }
 
    function toString() {
      return $this->lngZone . $this->latZone . " " . 
             $this->easting . " " . $this->northing;
    }
    
  }
  
  
  // ================================================================== RefEll
  
  class RefEll {
  
    var $maj;
    var $min;
    var $ecc;
  
    function RefEll($maj, $min) {
      $this->maj = $maj;
      $this->min = $min;
      $this->ecc = (($maj * $maj) - ($min * $min)) / ($maj * $maj);
    }
  
  }
  
  
  // ================================================== Mathematical Functions
  
  function sinSquared($x) {
    return sin($x) * sin($x);
  }
  
  function cosSquared($x) {
    return cos($x) * cos($x);
  }
  
  function tanSquared($x) {
    return tan($x) * tan($x);
  }
  
  function sec($x) {
    return 1.0 / cos($x);
  }
  

  // ========================================================= Other Functions
  
  function LatLngDistance($from, $to) {
    $er = 6366.707;
    
    $latFrom = deg2rad($from->lat);
    $latTo   = deg2rad($to->lat);
    $lngFrom = deg2rad($from->lng);
    $lngTo   = deg2rad($to->lng);
    
    $x1 = $er * cos($lngFrom) * sin($latFrom);
    $y1 = $er * sin($lngFrom) * sin($latFrom);
    $z1 = $er * cos($latFrom);

    $x2 = $er * cos($lngTo) * sin($latTo);
    $y2 = $er * sin($lngTo) * sin($latTo);
    $z2 = $er * cos($latTo);

    $d = sqrt(($x1-$x2)*($x1-$x2)+($y1-$y2)*($y1-$y2)+($z1-$z2)*($z1-$z2));
    
    return $d;
  }


  // ==================================================== Conversion Functions
  
  function OSRefToLatLng($ref) {
    $airy1830 = new RefEll(6377563.396, 6356256.909);
    $OSGB_F0  = 0.9996012717;
    $N0       = -100000.0;
    $E0       = 400000.0;
    $phi0     = deg2rad(49.0);
    $lambda0  = deg2rad(-2.0);
    $a        = $airy1830->maj;
    $b        = $airy1830->min;
    $eSquared = $airy1830->ecc;
    $phi      = 0.0;
    $lambda   = 0.0;
    $E        = $ref->easting;
    $N        = $ref->northing;
    $n        = ($a - $b) / ($a + $b);
    $M        = 0.0;
    $phiPrime = (($N - $N0) / ($a * $OSGB_F0)) + $phi0;
    do {
      $M =
        ($b * $OSGB_F0)
          * (((1 + $n + ((5.0 / 4.0) * $n * $n) + ((5.0 / 4.0) * $n * $n * $n))
            * ($phiPrime - $phi0))
            - (((3 * $n) + (3 * $n * $n) + ((21.0 / 8.0) * $n * $n * $n))
              * sin($phiPrime - $phi0)
              * cos($phiPrime + $phi0))
            + ((((15.0 / 8.0) * $n * $n) + ((15.0 / 8.0) * $n * $n * $n))
              * sin(2.0 * ($phiPrime - $phi0))
              * cos(2.0 * ($phiPrime + $phi0)))
            - (((35.0 / 24.0) * $n * $n * $n)
              * sin(3.0 * ($phiPrime - $phi0))
              * cos(3.0 * ($phiPrime + $phi0))));
      $phiPrime += ($N - $N0 - $M) / ($a * $OSGB_F0);
    } while (($N - $N0 - $M) >= 0.001);
    $v = $a * $OSGB_F0 * pow(1.0 - $eSquared * sinSquared($phiPrime), -0.5);
    $rho =
      $a
        * $OSGB_F0
        * (1.0 - $eSquared)
        * pow(1.0 - $eSquared * sinSquared($phiPrime), -1.5);
    $etaSquared = ($v / $rho) - 1.0;
    $VII = tan($phiPrime) / (2 * $rho * $v);
    $VIII =
      (tan($phiPrime) / (24.0 * $rho * pow($v, 3.0)))
        * (5.0
          + (3.0 * tanSquared($phiPrime))
          + $etaSquared
          - (9.0 * tanSquared($phiPrime) * $etaSquared));
    $IX =
      (tan($phiPrime) / (720.0 * $rho * pow($v, 5.0)))
        * (61.0
          + (90.0 * tanSquared($phiPrime))
          + (45.0 * tanSquared($phiPrime) * tanSquared($phiPrime)));
    $X = sec($phiPrime) / $v;
    $XI =
      (sec($phiPrime) / (6.0 * $v * $v * $v))
        * (($v / $rho) + (2 * tanSquared($phiPrime)));
    $XII =
      (sec($phiPrime) / (120.0 * pow($v, 5.0)))
        * (5.0
          + (28.0 * tanSquared($phiPrime))
          + (24.0 * tanSquared($phiPrime) * tanSquared($phiPrime)));
    $XIIA =
      (sec($phiPrime) / (5040.0 * pow($v, 7.0)))
        * (61.0
          + (662.0 * tanSquared($phiPrime))
          + (1320.0 * tanSquared($phiPrime) * tanSquared($phiPrime))
          + (720.0
            * tanSquared($phiPrime)
            * tanSquared($phiPrime)
            * tanSquared($phiPrime)));
    $phi =
      $phiPrime
        - ($VII * pow($E - $E0, 2.0))
        + ($VIII * pow($E - $E0, 4.0))
        - ($IX * pow($E - $E0, 6.0));
    $lambda =
      $lambda0
        + ($X * ($E - $E0))
        - ($XI * pow($E - $E0, 3.0))
        + ($XII * pow($E - $E0, 5.0))
        - ($XIIA * pow($E - $E0, 7.0));

    return new LatLng(rad2deg($phi), rad2deg($lambda));
  }
  
  
  /**
   * Convert a latitude and longitude into an OSGB grid reference
   * 
   * @param latitudeLongitude the latitude and longitude to convert
   * @return the OSGB grid reference
   * @since 0.1
   */
  function LatLngToOSRef($latLng) {
    $airy1830 = new RefEll(6377563.396, 6356256.909);
    $OSGB_F0  = 0.9996012717;
    $N0       = -100000.0;
    $E0       = 400000.0;
    $phi0     = deg2rad(49.0);
    $lambda0  = deg2rad(-2.0);
    $a        = $airy1830->maj;
    $b        = $airy1830->min;
    $eSquared = $airy1830->ecc;
    $phi = deg2rad($latLng->lat);
    $lambda = deg2rad($latLng->lng);
    $E = 0.0;
    $N = 0.0;
    $n = ($a - $b) / ($a + $b);
    $v = $a * $OSGB_F0 * pow(1.0 - $eSquared * sinSquared($phi), -0.5);
    $rho =
      $a * $OSGB_F0 * (1.0 - $eSquared) * pow(1.0 - $eSquared * sinSquared($phi), -1.5);
    $etaSquared = ($v / $rho) - 1.0;
    $M =
      ($b * $OSGB_F0)
        * (((1 + $n + ((5.0 / 4.0) * $n * $n) + ((5.0 / 4.0) * $n * $n * $n))
          * ($phi - $phi0))
          - (((3 * $n) + (3 * $n * $n) + ((21.0 / 8.0) * $n * $n * $n))
            * sin($phi - $phi0)
            * cos($phi + $phi0))
          + ((((15.0 / 8.0) * $n * $n) + ((15.0 / 8.0) * $n * $n * $n))
            * sin(2.0 * ($phi - $phi0))
            * cos(2.0 * ($phi + $phi0)))
          - (((35.0 / 24.0) * $n * $n * $n)
            * sin(3.0 * ($phi - $phi0))
            * cos(3.0 * ($phi + $phi0))));
    $I = $M + $N0;
    $II = ($v / 2.0) * sin($phi) * cos($phi);
    $III =
      ($v / 24.0)
        * sin($phi)
        * pow(cos($phi), 3.0)
        * (5.0 - tanSquared($phi) + (9.0 * $etaSquared));
    $IIIA =
      ($v / 720.0)
        * sin($phi)
        * pow(cos($phi), 5.0)
        * (61.0 - (58.0 * tanSquared($phi)) + pow(tan($phi), 4.0));
    $IV = $v * cos($phi);
    $V = ($v / 6.0) * pow(cos($phi), 3.0) * (($v / $rho) - tanSquared($phi));
    $VI =
      ($v / 120.0)
        * pow(cos($phi), 5.0)
        * (5.0
          - (18.0 * tanSquared($phi))
          + (pow(tan($phi), 4.0))
          + (14 * $etaSquared)
          - (58 * tanSquared($phi) * $etaSquared));

    $N =
      $I
        + ($II * pow($lambda - $lambda0, 2.0))
        + ($III * pow($lambda - $lambda0, 4.0))
        + ($IIIA * pow($lambda - $lambda0, 6.0));
    $E =
      $E0
        + ($IV * ($lambda - $lambda0))
        + ($V * pow($lambda - $lambda0, 3.0))
        + ($VI * pow($lambda - $lambda0, 5.0));

    return new OSRef($E, $N);
  }
  
  
  /**
   * Convert an UTM reference to a latitude and longitude
   * 
   * @param ellipsoid A reference ellipsoid to use
   * @param utm the UTM reference to convert
   * @return the converted latitude and longitude
   * @since 0.2
   */
  function UTMRefToLatLng($utm) {

    $wgs84 = new RefEll(6378137, 6356752.314);
    $UTM_F0   = 0.9996;
    $a = $wgs84->maj;
    $eSquared = $wgs84->ecc;
    $ePrimeSquared = $eSquared / (1.0 - $eSquared);
    $e1 = (1 - sqrt(1 - $eSquared)) / (1 + sqrt(1 - $eSquared));
    $x = $utm->easting - 500000.0;;
    $y = $utm->northing;
    $zoneNumber = $utm->lngZone;
    $zoneLetter = $utm->latZone;
    
    $longitudeOrigin = ($zoneNumber - 1.0) * 6.0 - 180.0 + 3.0;
    
    // Correct y for southern hemisphere
    if ((ord($zoneLetter) - ord("N")) < 0) {
      $y -= 10000000.0;
    }

    $m = $y / $UTM_F0;
    $mu =
      $m
        / ($a
          * (1.0
            - $eSquared / 4.0
            - 3.0 * $eSquared * $eSquared / 64.0
            - 5.0
              * pow($eSquared, 3.0)
              / 256.0));

    $phi1Rad =
      $mu
        + (3.0 * $e1 / 2.0 - 27.0 * pow($e1, 3.0) / 32.0) * sin(2.0 * $mu)
        + (21.0 * $e1 * $e1 / 16.0 - 55.0 * pow($e1, 4.0) / 32.0)
          * sin(4.0 * $mu)
        + (151.0 * pow($e1, 3.0) / 96.0) * sin(6.0 * $mu);

    $n =
      $a
        / sqrt(1.0 - $eSquared * sin($phi1Rad) * sin($phi1Rad));
    $t = tan($phi1Rad) * tan($phi1Rad);
    $c = $ePrimeSquared * cos($phi1Rad) * cos($phi1Rad);
    $r =
      $a
        * (1.0 - $eSquared)
        / pow(
          1.0 - $eSquared * sin($phi1Rad) * sin($phi1Rad),
          1.5);
    $d = $x / ($n * $UTM_F0);

    $latitude = (
    
      $phi1Rad
        - ($n * tan($phi1Rad) / $r)
          * ($d * $d / 2.0
            - (5.0
              + (3.0 * $t)
              + (10.0 * $c)
              - (4.0 * $c * $c)
              - (9.0 * $ePrimeSquared))
              * pow($d, 4.0)
              / 24.0
            + (61.0
              + (90.0 * $t)
              + (298.0 * $c)
              + (45.0 * $t * $t)
              - (252.0 * $ePrimeSquared)
              - (3.0 * $c * $c))
              * pow($d, 6.0)
              / 720.0)) * (180.0 / pi());

    $longitude = $longitudeOrigin + (
      ($d
        - (1.0 + 2.0 * $t + $c) * pow($d, 3.0) / 6.0
        + (5.0
          - (2.0 * $c)
          + (28.0 * $t)
          - (3.0 * $c * $c)
          + (8.0 * $ePrimeSquared)
          + (24.0 * $t * $t))
          * pow($d, 5.0)
          / 120.0)
        / cos($phi1Rad)) * (180.0 / pi());
    
    return new LatLng($latitude, $longitude);
  }
  
  
  /**
   * Convert a latitude and longitude to an UTM reference 
   * 
   * @param ellipsoid A reference ellipsoid to use
   * @param latitudeLongitude The latitude and longitude to convert
   * @return the converted UTM reference
   * @since 0.2
   */
  function latLngToUTMRef($latLng) {

    $wgs84 = new RefEll(6378137, 6356752.314);
    $UTM_F0   = 0.9996;
    $a = $wgs84->maj;
    $eSquared = $wgs84->ecc;
    $longitude = $latLng->lng;
    $latitude = $latLng->lat;
    
    $latitudeRad = $latitude * (pi() / 180.0);
    $longitudeRad = $longitude * (pi() / 180.0);
    $longitudeZone = (int) (($longitude + 180.0) / 6.0) + 1;

    // Special zone for Norway
    if ($latitude >= 56.0
      && $latitude < 64.0
      && $longitude >= 3.0
      && $longitude < 12.0) {
      $longitudeZone = 32;
    }

    // Special zones for Svalbard
    if ($latitude >= 72.0 && $latitude < 84.0) {
      if ($longitude >= 0.0 && $longitude < 9.0) {
        $longitudeZone = 31;
      } else if ($longitude >= 9.0 && $longitude < 21.0) {
        $longitudeZone = 33;
      } else if ($longitude >= 21.0 && $longitude < 33.0) {
        $longitudeZone = 35;
      } else if ($longitude >= 33.0 && $longitude < 42.0) {
        $longitudeZone = 37;
      }
    }
    
    $longitudeOrigin = ($longitudeZone - 1) * 6 - 180 + 3;
    $longitudeOriginRad = $longitudeOrigin * (pi() / 180.0);

    $UTMZone = getUTMLatitudeZoneLetter($latitude);

    $ePrimeSquared = ($eSquared) / (1 - $eSquared);

    $n = $a / sqrt(1 - $eSquared * sin($latitudeRad) * sin($latitudeRad));
    $t = tan($latitudeRad) * tan($latitudeRad);
    $c = $ePrimeSquared * cos($latitudeRad) * cos($latitudeRad);
    $A = cos($latitudeRad) * ($longitudeRad - $longitudeOriginRad);

    $M =
      $a
        * ((1
          - $eSquared / 4
          - 3 * $eSquared * $eSquared / 64
          - 5 * $eSquared * $eSquared * $eSquared / 256)
          * $latitudeRad
          - (3 * $eSquared / 8
            + 3 * $eSquared * $eSquared / 32
            + 45 * $eSquared * $eSquared * $eSquared / 1024)
            * sin(2 * $latitudeRad)
          + (15 * $eSquared * $eSquared / 256
            + 45 * $eSquared * $eSquared * $eSquared / 1024)
            * sin(4 * $latitudeRad)
          - (35 * $eSquared * $eSquared * $eSquared / 3072)
            * sin(6 * $latitudeRad));

    $UTMEasting =
      (double) ($UTM_F0
        * $n
        * ($A
          + (1 - $t + $c) * pow($A, 3.0) / 6
          + (5 - 18 * $t + $t * $t + 72 * $c - 58 * $ePrimeSquared)
            * pow($A, 5.0)
            / 120)
        + 500000.0);

    $UTMNorthing =
      (double) ($UTM_F0
        * ($M
          + $n
            * tan($latitudeRad)
            * ($A * $A / 2
              + (5 - $t + (9 * $c) + (4 * $c * $c)) * pow($A, 4.0) / 24
              + (61 - (58 * $t) + ($t * $t) + (600 * $c) - (330 * $ePrimeSquared))
                * pow($A, 6.0)
                / 720)));
    
    // Adjust for the southern hemisphere
    if ($latitude < 0) {
      $UTMNorthing += 10000000.0;
    }

    return new UTMRef($UTMEasting, $UTMNorthing, $UTMZone, $longitudeZone);
  }
  
  
  /**
   *  Work out the UTM latitude zone from the latitude
   * 
   * @param latitude
   * @return
   * @since 0.2
   */
  function getUTMLatitudeZoneLetter($latitude) {
    if ((84 >= $latitude) && ($latitude >= 72)) return "X";
    else if (( 72 > $latitude) && ($latitude >=  64)) return "W";
    else if (( 64 > $latitude) && ($latitude >=  56)) return "V";
    else if (( 56 > $latitude) && ($latitude >=  48)) return "U";
    else if (( 48 > $latitude) && ($latitude >=  40)) return "T";
    else if (( 40 > $latitude) && ($latitude >=  32)) return "S";
    else if (( 32 > $latitude) && ($latitude >=  24)) return "R";
    else if (( 24 > $latitude) && ($latitude >=  16)) return "Q";
    else if (( 16 > $latitude) && ($latitude >=   8)) return "P";
    else if ((  8 > $latitude) && ($latitude >=   0)) return "N";
    else if ((  0 > $latitude) && ($latitude >=  -8)) return "M";
    else if (( -8 > $latitude) && ($latitude >= -16)) return "L";
    else if ((-16 > $latitude) && ($latitude >= -24)) return "K";
    else if ((-24 > $latitude) && ($latitude >= -32)) return "J";
    else if ((-32 > $latitude) && ($latitude >= -40)) return "H";
    else if ((-40 > $latitude) && ($latitude >= -48)) return "G";
    else if ((-48 > $latitude) && ($latitude >= -56)) return "F";
    else if ((-56 > $latitude) && ($latitude >= -64)) return "E";
    else if ((-64 > $latitude) && ($latitude >= -72)) return "D";
    else if ((-72 > $latitude) && ($latitude >= -80)) return "C";
    else return 'Z';
  }
?>